'''
filehanding.py: this module provides classes for passing files between
processes on distributed computing platforms that may not share a common
file system.
'''
from __future__ import print_function

import os
import tempfile
import zlib
from shutil import copyfile

'''
This module defines classes to handle files in distributed environments
where filesyatems may not be shared.

Objects of each class are instantiated with the path of an existing file on
an existing file syatem:

    fh = FileHandle('/path/to/file')

and have a save() method that creates a local copy of that file:

    filename_here = fh.save(filename_here)

and also an as_file() method which returns a path that points at a (maybe temporary) copy of the file contents, suitable for use in cases where a function requires a filename:

    with open(fh.as_file()) as f:
        ...

'''
class FileHandle(object):
    '''
    Base class for file handlers
    '''
    def __init__(self, path, session_dir=None):
        self.path = os.path.abspath(path)
        self.session_dir = session_dir
   
    def __str__(self):
        return "Filehandle for file {}".format(self.path)

    def save(self, path):
        """
        Save the file

        args:
            path (str): file path

        returns:
            str: the path
        """
        abspath = os.path.abspath(path)
        if abspath != self.path:
            copyfile(self.path, path)
        return path

    def as_file(self):
        """
        Returns a path which points at the file
        """
        return self.path

class TempFileHandle(FileHandle):
    '''
    File handler that used $TMPDIR as shared space
    '''
    def __init__(self, path, session_dir=None):
        super(TempFileHandle, self).__init__(path, session_dir)
        ext = os.path.splitext(path)[1]
        tmpdir = os.path.join(os.path.dirname(tempfile.mkdtemp()), session_dir)
        if not os.path.exists(tmpdir):
            os.mkdir(tmpdir)
        self.tmp_path = tempfile.NamedTemporaryFile(dir=tmpdir, suffix=ext, delete=False).name
        copyfile(self.path, self.tmp_path)
        
    def save(self, path):
        """
        Save a copy of the file.

        args:
            path (str): the path to use

        returns:
            str: the path used
        """
        copyfile(self.tmp_path, path)
        return path

    def as_file(self):
        """
        returns a path that points at the file
        """
        return self.tmp_path

class SharedFileHandle(FileHandle):
    '''
    File handler that stores data in some shared - e.g. NFS - space.
    Methods allow for $SHARED to point at different places on different nodes
    '''
    def __init__(self, path, session_dir=None):
        super(SharedFileHandle, self).__init__(path, session_dir)
        shared_dir = os.getenv('SHARED')
        if shared_dir is None:
            raise IOError('Error - environment variable $SHARED is not set')
        shared_dir = os.path.join(shared_dir, self.session_dir)
        if not os.path.exists(shared_dir):
            os.mkdir(shared_dir)
        ext = os.path.splitext(path)[1]
        self.shared_path = tempfile.NamedTemporaryFile(suffix=ext,
                                                       dir=shared_dir,
                                                       delete=False).name
        copyfile(self.path, self.shared_path)

    def save(self, path):
        """
        Save a copy of the file

        args:
            path (str): path for the saved file

        returns:
            str: path of the saved file
        """
        if self.session_dir is None:
            raise SystemError('Error - session_dir not set')
        shared_path = self.as_file()
        copyfile(shared_path, path)
        return path

    def as_file(self):
        """
        Returns a path that points at the file
        """
        shared_dir = os.getenv('SHARED')
        if shared_dir is None:
            raise IOError('Error - environment variable $SHARED is not set')
        shared_dir = os.path.join(shared_dir, self.session_dir)
        shared_path = os.path.join(shared_dir, os.path.basename(self.shared_path))
        return shared_path

class CompressedFileHandle(FileHandle):
    '''
    File handler that stores data in memory
    '''
    def __init__(self, path, session_dir=None):
        super(CompressedFileHandle, self).__init__(path)
        with open(self.path, 'rb') as f:
            self.compressed_data = zlib.compress(f.read())

    def save(self, path):
        """
        Save a copy of the file

        args:
            path (str): path for the saved file

        returns:
            str: path of the saved file
        """
        with open(path, 'wb') as f:
            f.write(zlib.decompress(self.compressed_data))
        return path

    def as_file(self):
        """
        Returns a path that points at the file
        """
        ext = os.path.splitext(self.path)[1]
        tmp_path = tempfile.NamedTemporaryFile(suffix=ext, delete=False).name
        return self.save(tmp_path)



'''
Xflowlib: the Xflow library to build workflows on an Xbow cluster
'''
import re
import subprocess
import os
import tempfile
import shutil
import copy
import hashlib
import glob
import uuid
from path import Path
from .filehandling import SharedFileHandle, CompressedFileHandle, TempFileHandle

filehandler = None
filehandler_type = None
session_dir = str(uuid.uuid4())

def set_filehandler(fh_type):
    """
    Set the type of file handler that will be used to pass file data between
    kernels.

    Currently three options are available. 

    The most basic ('tmp') uses the
    temporary file system on the host for sharing files. Obviously this only
    works if all workers reside on the same host.

    The second option is 'shared' which uses some shared (e.g. NFS) file
    system to pass files between worker nodes. This requires each node to
    have an environment variable SHARED which points to the mount point of
    the shared file system on that node.

    The third option is 'memory' which doesn't use a file system at all, 
    file data is stored in memory and passed to workers the same way as 
    all other data objects.

    args:
        fh_type (str): one of 'tmp', 'shared', or 'memory'
    """
    global filehandler_type
    global filehandler
    fh_types = ['tmp', 'shared', 'memory']
    fh_list = [TempFileHandle, SharedFileHandle, CompressedFileHandle]
    if not fh_type in fh_types:
        raise ValueError('Error - argument must be one of "tmp", "shared". '
                         'or "memory"')
    filehandler_type = fh_type
    filehandler = fh_list[fh_types.index(fh_type)]

def purge():
    '''
    Remove all temporary files for the current session.
    '''
    if filehandler_type == 'tmp':
        tmpdir = os.path.join(os.path.dirname(tempfile.mkdtemp()), session_dir)
    elif filehandler_type == 'shared':
        tmpdir = os.path.join(os.getenv('SHARED'), session_dir)
    else:
        tmpdir = None
    if tmpdir is not None:
        try:
            shutil.rmtree(tmpdir)
        except:
            pass

def load(filename):
    '''
    Returns a FileHandle for a path
    '''
    if filehandler is None:
        set_filehandler('memory')
    return filehandler(filename, session_dir=session_dir)

class Filepack(object):
    """
    A collection of files
    """
    def __init__(self, filelist):
        self.filepack = {}
        for filename in filelist:
            self.filepack[filename] = filehandler(filename, session_dir=session_dir)

    def unpack(self, outputdir='.'):
        '''
        Unpack the files in the Filepack into the given directory.
        '''
        for filename in self.filepack:
            outname = os.path.join(outputdir, filename)
            self.filepack[filename].save(outname)

class SubprocessKernel(object):
    '''
    A kernel that runs a command-line executable
    '''
    def __init__(self, template):
        """
        Arguments:
            template (str): a template for the command to be executed
        """
        self.template = template
        self.inputs = []
        self.outputs = []
        self.constants = []
        self.STDOUT = None
        if filehandler is None:
            set_filehandler('memory')
        self.filehandler = filehandler
        if session_dir is None:
            raise SystemError('Error - session_dir is not set')
        self.session_dir = session_dir

        self.variables = []
        for key in re.findall(r'\{.*?\}', self.template):
            self.variables.append(key[1:-1])

    def set_inputs(self, inputs):
        """
        Set the inputs the kernel requires
        """
        if not isinstance(inputs, list):
            raise TypeError('Error - inputs must be of type list,'
                    ' not of type {}'.format(type(inputs)))
        self.inputs = inputs

    def set_outputs(self, outputs):
        """
        Set the outputs the kernel produces
        """
        if not isinstance(outputs, list):
            raise TypeError('Error - outputs must be of type list,'
                    ' not of type {}'.format(type(outputs)))
        self.outputs = outputs

    def set_constant(self, key, value):
        """
        Set a constant for the kernel
        If it was previously defined as an input variable, remove it from
        that list.
        """
        d = {}
        d['name'] = key
        if isinstance(value, str):
            if os.path.exists(value):
                d['value'] = self.filehandler(value, session_dir=self.session_dir)
            else:
                d['value'] = value
        else:
            d['value'] = value
        self.constants.append(d)

        if key in self.inputs:
            self.inputs.remove(key)

    def copy(self):
        '''
        Return a copy of the kernel
        '''
        return copy.deepcopy(self)

    def run(self, *args):
        """
        Run the kernel with the given inputs.
        Args:
            args: positional arguments whose order should match self.inputs

        Returns:
            tuple : outputs in the order they appear in
                self.outputs
        """
        outputs = []
        td = tempfile.mkdtemp()
        with Path(td) as tmpdir:
            var_dict = {}
            for i in range(len(args)):
                if self.inputs[i] in self.variables:
                    var_dict[self.inputs[i]] = args[i]
                else:
                    try:
                        args[i].save(self.inputs[i])
                    except AttributeError:
                        raise TypeError('Error with variable {} {}'.format(i, args[i]))
            for d in self.constants:
                try:
                    d['value'].save(d['name'])
                except AttributeError:
                    var_dict[d['name']] = d['value']
            cmd = self.template.format(**var_dict)
            try:
                result = subprocess.run(cmd, shell=True,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE,
                                        check=True)
            except subprocess.CalledProcessError as e:
                result = CalledProcessError(e)
                if not 'RESULT' in self.outputs:
                    raise result

            self.STDOUT = result.stdout.decode()
            for outfile in self.outputs:
                if os.path.exists(outfile):
                    outputs.append(self.filehandler(outfile, session_dir=self.session_dir))
                else:
                    if outfile == 'STDOUT':
                        outputs.append(self.STDOUT)
                    elif outfile == 'RESULT':
                        outputs.append(result)
                    else:
                        outputs.append(None)
        try:
            shutil.rmtree(td)
        except:
            pass
        if len(outputs) == 1:
            outputs = outputs[0]
        else:
            outputs = tuple(outputs)
        return outputs
     
class FunctionKernel(object):
    def __init__(self, func):
        """
        Arguments:
            func: the Python function to wrap
            filetype (FileHandle): the file-type handler
        """
        self.func = func
        self.inputs = []
        self.outputs = []
        self.constants = {}
        self.tmpdir = None
        if filehandler is None:
            set_filehandler('memory')
        self.filehandler = filehandler
        self.session_dir = session_dir

    def set_inputs(self, inputs):
        """
        Set the inputs the kernel requires
        """
        self.inputs = inputs

    def set_outputs(self, outputs):
        """
        Set the outputs the kernel produces
        """
        self.outputs = outputs

    def set_constant(self, key, value):
        """
        Set a parameters for the kernel
        """
        self.constants[key] = value
        if isinstance(value, str):
            if os.path.exists(value):
                self.constants[key] = self.filehandler(value, session_dir=self.session_dir)

    def copy(self):
        """
        Return a copy of the kernel
        """
        return copy.copy(self)

    def run(self, *args):
        """
        Run the kernel/function with the given arguments.

        Returns:
            Whatever the function returns, with output files converted
                to FileHandle objects
        """
        #td = tempfile.TemporaryDirectory(dir=self.tmpdir)
        #with Path(td.name) as tmpdir:
        td = tempfile.mkdtemp(dir=self.tmpdir)
        with Path(td) as tmpdir:
            indict = {}
            for i, v in enumerate(args):
                if isinstance(v, dict):
                    for k in v:
                        if k in self.inputs:
                            try:
                                indict[k] = v[k].save(os.path.basename(v[k].path))
                            except AttributeError:
                                indict[k] = v[k]
                else:
                    try:
                        indict[self.inputs[i]] = v.save(os.path.basename(v.path))
                    except AttributeError:
                        indict[self.inputs[i]] = v
            for k in self.constants:
                try:
                    indict[k] = self.constants[k].save(os.path.basename(self.constants[k].path))
                except AttributeError:
                    indict[k] = self.constants[k]
            result = self.func(**indict)
            if not isinstance(result, list):
                result = [result]
            outputs = []
            for i, v in enumerate(result):
                if isinstance(v, str):
                    if os.path.exists(v):
                        outputs.append(self.filehandler(v, session_dir=self.session_dir))
                    else:
                        outputs.append(v)
                else:
                    outputs.append(v)
        try:
            shutil.rmtree(td)
        except:
            pass
        if len(outputs) == 1:
            outputs = outputs[0]
        else:
            outputs = tuple(outputs)
        return outputs
class XflowError(Exception):
    """
    Base class for Xflowlib exceptions.
    """
    pass

class CalledProcessError(XflowError):
    """
    Exception raised if a kernel fails with an error.

    A cosmetic wrapper round subprocess.CalledProcessError
    """

    def __init__(self, e):
        self.cmd = e.cmd
        self.returncode = e.returncode
        self.stdout = e.stdout
        self.stderr = e.stderr
        self.output = self.stdout

    def __str__(self):
        return 'Error: command "{}" failed with return code {}; STDOUT="{}"; STDERR="{}"'.format(self.cmd, self.returncode, self.stdout, self.stderr)



'''
Clients.py: thin wrapper over dask client
'''
from __future__ import print_function
import socket
import hashlib
import subprocess
import glob
import os
from dask.distributed import Client, LocalCluster
from .xflowlib import FunctionKernel, SubprocessKernel

def dask_client(scheduler_file=None, local=False, port=8786):
    """
    returns an instance of a dask.distributed client
    """
    if local:
        cluster = LocalCluster()
        client = Client(cluster)
    elif scheduler_file:
        client = Client(scheduler_file=scheduler_file)
    else:
        ip_address = socket.gethostbyname(socket.gethostname())
        dask_scheduler = '{}:{}'.format(ip_address, port)
        try:
            client = Client(dask_scheduler, timeout=5)
        except IOError:
            print('Warning: using local dask client')
            cluster = LocalCluster()
            client = Client(cluster)
    return client

class XflowClient(object):
    '''Thin wrapper around Dask client so functions that return multiple
       values (tuples) generate tuples of futures rather than single futures.
    '''
    def __init__(self, **kwargs):
        self.tmpdir = kwargs.pop('tmpdir', None)
        self.client = dask_client(**kwargs)

    def cluster(self):
        """
        Basic info about the cluster
        """
        return self.client.cluster

    def close(self):
        """
        The close() method of the underlying dask client
        """
        return self.client.close

    def upload(self, some_object):
        """
        Upload some data/object to the Xbow cluster.

        args:
            fsome_object (any type): what to upload

        returns:
            dask.Future
        """
        return self.client.scatter(some_object, broadcast=True)

    def unpack(self, kernel, future):
        """
        Unpacks the single future returned by kernel when run through
        a dask submit() or map() method, returning a tuple of futures.

        The outputs attribute of kernel lists how many values kernel
        should properly return.

        args:
            kernel (Kernel): the kernel that generated the future
            future (Future): the future returned by kernel

        returns:
            future or tuple of futures.
        """
        if len(kernel.outputs) == 1:
            return future
        outputs = []
        for i in range(len(kernel.outputs)):
            outputs.append(self.client.submit(lambda tup, j: tup[j], future, i))
        return tuple(outputs)

    def submit(self, func, *args):
        """
        Wrapper round the dask submit() method, so that a tuple of
        futures, rather than just one future, is returned.

        args:
            func (function/kernel): the function to be run
            args (list): the function arguments
        returns:
            future or tuple of futures
        """
        if isinstance(func, SubprocessKernel):
            func.tmpdir = self.tmpdir
            future = self.client.submit(func.run, *args, pure=False)
            return self.unpack(func, future)
        if isinstance(func, FunctionKernel):
            func.tmpdir = self.tmpdir
            future = self.client.submit(func.run, *args, pure=False)
            return self.unpack(func, future)
        else:
            return self.client.submit(func, *args)

    def _lt2tl(self, l):
        '''converts a list of tuples to a tuple of lists'''
        result = []
        for i in range(len(l[0])):
            result.append([t[i] for t in l])
        return tuple(result)

    def map(self, func, *iterables):
        """
        Wrapper arounf the dask map() method so it returns lists of
        tuples of futures, rather than lists of futures.

        args:
            func (function): the function to be mapped
            iterables (iterables): the function arguments

        returns:
            list or tuple of lists: futures returned by the mapped function
        """
        its = []
        maxlen = 0
        for iterable in iterables:
            if isinstance(iterable, list):
                l = len(iterable)
                if l > maxlen:
                    maxlen = l
        for iterable in iterables:
            if isinstance(iterable, list):
                l = len(iterable)
                if l != maxlen:
                    raise ValueError('Error: not all iterables are same length')
                its.append(iterable)
            else:
                its.append([iterable] * maxlen)
        if isinstance(func, SubprocessKernel):
            func.tmpdir = self.tmpdir
            futures = self.client.map(func.run, *its, pure=False)
            result = [self.unpack(func, future) for future in futures]
        elif isinstance(func, FunctionKernel):
            func.tmpdir = self.tmpdir
            futures = self.client.map(func.run, *its, pure=False)
            result = [self.unpack(func, future) for future in futures]
        else:
            result =  self.client.map(func, *its, pure=False)
        if isinstance(result[0], tuple):
            result = self._lt2tl(result)
        return result

    def execall(self, cmd):
        '''
        Run a command on all workers in a cluster.
        '''
        full_cmd = 'cd {}; {}'.format(os.getcwd(), cmd)
        def myfunc(cmd):
            try:
                r = subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True)
                returncode = 0
            except subprocess.CalledProcessError as e:
                r = e.output
                returncode = e.returncode
            return {'returncode': returncode, 'output': r}
        result = self.client.run(myfunc, full_cmd)
        return result

    def install(self, package, sudo=False):
        '''
        Install a package on all workers in a cluster.
        '''
        if sudo:
            cmd = 'sudo pip install {}'.format(package)
        else:
            cmd = 'pip install {}'.format(package)

        result = self.execall(cmd)
        errors = False
        errortext = ''
        for key in result.keys():
            if result[key]['returncode'] != 0:
                errortext += 'Warning: install failed for worker {}\n'.format(key)
                errortext += result[key]['output'].decode('utf-8')
                errors = True
        if errors:
            raise RuntimeError(errortext)

def md5checksum(filename):
    """
    Returns the md5 checksum of a file
    """
    return hashlib.md5(open(filename, 'rb').read()).hexdigest()

def unpack_run_and_pack(cmd, filepack, tmpdir=None):
    '''
    Unpack some data files, run a command, return the results.
    '''
    with tempfile.TemporaryDirectory(dir=tmpdir) as tmpd:
        os.chdir(tmpd)
        filepack.unpack()
        m5s = {}
        for filename in filepack.filepack:
            m5s[filename] = md5checksum(filename)

        try:
            result = subprocess.check_output(cmd, shell=True,
                                             stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            result = e.output
        with open('STDOUT', 'wb') as f:
            f.write(result)
        filelist = glob.glob('*')
        outfiles = []
        for filename in filelist:
            if not filename in filepack.filepack:
                outfiles.append(filename)
            elif md5checksum(filename) != m5s[filename]:
                outfiles.append(filename)
        outfilepack = Filepack(outfiles)
    return outfilepack

def remote_run(client, cmd, tmpdir=None):
    '''
    Run a command on a client, including file staging
    '''
    files = [f for f in glob.glob('*') if os.path.isfile(f)]
    filepack = client.scatter(Filepack(files))
    result = client.submit(unpack_run_and_pack, cmd, filepack, tmpdir)
    result.result().unpack()



from xbowflow import xflowlib
from xbowflow.clients import XflowClient
import subprocess

ls = xflowlib.SubprocessKernel('ls {x} > output')
ls.set_inputs(['x'])
#ls.set_outputs(['output', 'RESULT'])
ls.set_outputs(['output'])

if __name__ == '__main__':
    client = XflowClient(local=True)
    testdata = '-a'
    #output, result = client.submit(ls, testdata)
    output = client.submit(ls, testdata)
    #print(result.result())
    output.result().save('ls-result1.log')
    testdata = 'crap'
    #output, result = client.submit(ls, testdata)
    output = client.submit(ls, testdata)
    #print(result.result())
    output.result().save('ls-result2.log')
    client.close()



import unittest
import os

from xbowflow import xflowlib, filehandling

class TestTempFilehandlingMethods(unittest.TestCase):

    def test_filehandle_methods_for_temp(self):
        xflowlib.set_filehandler('tmp')
        self.assertEqual(xflowlib.filehandler_type, 'tmp')
        self.assertEqual(xflowlib.filehandler, filehandling.TempFileHandle)
        fh = xflowlib.load('data/test.txt')
        self.assertIsInstance(fh, filehandling.FileHandle)
        fh.save('tempfile.txt')
        self.assertTrue(os.path.exists('tempfile.txt'))
        with open('data/test.txt', 'rb') as f1:
            d1 = f1.read()
        with open('tempfile.txt', 'rb') as f2:
            d2 = f2.read()
        self.assertEqual(d1, d2)
        os.remove('tempfile.txt')
        self.assertFalse(os.path.exists('tempfile.txt'))
        tmpname = fh.as_file()
        self.assertTrue(os.path.exists(tmpname))
        #del(fh)
        #self.assertFalse(os.path.exists(tmpname))

    def test_filehandle_methods_for_shared(self):
        xflowlib.set_filehandler('shared')
        self.assertEqual(xflowlib.filehandler_type, 'shared')
        self.assertEqual(xflowlib.filehandler, filehandling.SharedFileHandle)
        fh = xflowlib.load('data/test.txt')
        self.assertIsInstance(fh, filehandling.FileHandle)
        fh.save('tempfile.txt')
        self.assertTrue(os.path.exists('tempfile.txt'))
        with open('data/test.txt', 'rb') as f1:
            d1 = f1.read()
        with open('tempfile.txt', 'rb') as f2:
            d2 = f2.read()
        self.assertEqual(d1, d2)
        os.remove('tempfile.txt')
        self.assertFalse(os.path.exists('tempfile.txt'))
        tmpname = fh.as_file()
        self.assertTrue(os.path.exists(tmpname))
        #del(fh)
        #self.assertFalse(os.path.exists(tmpname))

    def test_filehandle_methods_for_memory(self):
        xflowlib.set_filehandler('memory')
        self.assertEqual(xflowlib.filehandler_type, 'memory')
        self.assertEqual(xflowlib.filehandler, filehandling.CompressedFileHandle)
        fh = xflowlib.load('data/test.txt')
        self.assertIsInstance(fh, filehandling.FileHandle)
        fh.save('tempfile.txt')
        self.assertTrue(os.path.exists('tempfile.txt'))
        with open('data/test.txt', 'rb') as f1:
            d1 = f1.read()
        with open('tempfile.txt', 'rb') as f2:
            d2 = f2.read()
        self.assertEqual(d1, d2)
        os.remove('tempfile.txt')
        self.assertFalse(os.path.exists('tempfile.txt'))
        tmpname = fh.as_file()
        self.assertTrue(os.path.exists(tmpname))
        del(fh)
        #self.assertFalse(os.path.exists(tmpname))

    def test_filehandle_methods_for_bad_value(self):
        with self.assertRaises(ValueError):
            xflowlib.set_filehandler('should_fail')
